@@ -2,7 +2,7 @@
  * A GNOME front end for LinuxConf
  *
  * Copyright (C) 1997 the Free Software Foundation
- * Copyright (C) 1998 Red Hat Software, Inc.
+ * Copyright (C) 1998, 1999 Red Hat Software, Inc.
  *
  * Authors: Miguel de Icaza (miguel@gnu.org)
  *          Michael K. Johnson <johnsonm@redhat.com>
@@ -301,6 +301,13 @@ widget_dump_tree(widget_info *w) {
 			io_write_message ("dump %s %s %s\n", w->path,
 				group->id, (char *) w->data);
 		}
+	case C_TREESUB: {
+		tree_item_data *item = w->data;
+		if (item->expand) {
+			static int idnum = 0;
+			char *tpath = parse_get_tree_path((command *)w);
+			io_write_message("dump %s t%d %s\n", w->id, idnum++, tpath);
+		}}
 	default:
 		/* not a printable item */
 		break;
@@ -317,6 +324,15 @@ report_action (GtkWidget *widget, gpointer data) {
 	io_write_message ("action %s %s\n", w->path, w->id);
 }
 
+static void
+close_action (GtkWidget *widget, gpointer data) {
+	widget_info *w = get_widget_info(widget);
+
+	if (!w) return;
+	widget_dump_tree(w);
+	io_write_message ("action %s X\n", w->path ? w->path : w->id );
+}
+
 static void
 button_dump_form (GtkWidget *widget, gpointer data) {
 	widget_info *w = get_widget_info(widget);
@@ -325,6 +341,14 @@ button_dump_form (GtkWidget *widget, gpointer data) {
 	report_action (widget, data);
 }
 
+static void
+subtree_expand (GtkWidget *widget, gpointer data) {
+	widget_info *w = get_widget_info(widget);
+	tree_item_data *item = w->data;
+
+	item->expand = data;
+}
+
 static void
 entry_enter_pressed (GtkWidget *widget, gpointer data) {
 	mainform_data *md;
@@ -924,11 +948,18 @@ create_treeitem(widget_info *w, GtkWidget *p) {
 		gtk_tree_append (GTK_TREE(p), treeitem);
 		gtk_tree_item_set_subtree(GTK_TREE_ITEM(treeitem), subtree);
 		create_menutree (w, subtree);
-		gtk_tree_item_expand(GTK_TREE_ITEM(treeitem));
+		if (item->expand)
+			gtk_tree_item_expand(GTK_TREE_ITEM(treeitem));
+		else
+			gtk_tree_item_collapse(GTK_TREE_ITEM(treeitem));
+		gtk_signal_connect(GTK_OBJECT(treeitem), "expand",
+			(GtkSignalFunc) subtree_expand, 1);
+		gtk_signal_connect(GTK_OBJECT(treeitem), "collapse",
+			(GtkSignalFunc) subtree_expand, 0);
 	} else {
 		gtk_tree_append (GTK_TREE(p), treeitem);
-		bind_widget_info(treeitem, w);
 	}
+	bind_widget_info(treeitem, w);
 	/* HACK to get around Gtk brokenness -- gtk_widget_show_all()
 	 * later should make this unnecessary.
 	 */
@@ -1029,6 +1060,12 @@ create_mainform(widget_info *w)
 		GtkWidget *vbox;
 
 		mainwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
+		/* special case the close button */
+		set_widget_info(mainwindow, w);
+		gtk_signal_connect(GTK_OBJECT(mainwindow), "destroy",
+			(GtkSignalFunc) close_action, 0);
+		gtk_signal_connect(GTK_OBJECT(mainwindow), "delete_event",
+			(GtkSignalFunc) close_action, 0);
 		vbox = gtk_vbox_new (0, 0);
 		gtk_container_add (GTK_CONTAINER(mainwindow), vbox);
 
@@ -1088,6 +1125,10 @@ create_mainform(widget_info *w)
 		 * or if treemenu widget has not yet been sighted
 		 */
 		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
+		gtk_signal_connect(GTK_OBJECT(window), "destroy",
+			(GtkSignalFunc) close_action, 0);
+		gtk_signal_connect(GTK_OBJECT(window), "delete_event",
+			(GtkSignalFunc) close_action, 0);
 		bind_widget_info(window, w);
 		table = create_table(w);  /* this creates the WHOLE tree */
 		gtk_window_set_title(GTK_WINDOW(window), w->str);
